proc changeSlashes { name } {
    global tcl_UNIX

    if !$tcl_UNIX {
        regsub -all \/ $name \\ cname
        return $cname
    } else {
        return $name
    }
}


#
# If 'vec' is nonempty and its first record contains less than
# 'fcount' fields, scan all records and add
# new fields from 'names' with the default values 'defaults'.
#
proc dataUprev { vec fcount names defaults {force 0} } {
	if {[$vec size] == 0} { return 0 }
	
	if { ! $force && [llength [$vec record 0]] == $fcount} {
		return 0
	}

	set extend [llength $names]
	for {set i 0} {$i < [$vec size]} {incr i} {
		set ofc [llength [$vec record $i]]
		if {$ofc + $extend < $fcount} {
			puts stderr "Warning: $vec record $i is short"
		}
		set here $ofc
		while {$here < $fcount} {
			set j [expr $here-($fcount-$extend)]
			if {$j < 0} {
				incr here
				continue
			}
			$vec at $i [lindex $names $j] [lindex $defaults $j]
			incr here
		}
	}
	return 1
}

proc dataFileRead { file vec args } {
    if [catch {set fid [open $file r]} result] {
        return 0
    }

    set opt [llength $args]

    if $opt {
        set delim [lindex $args 0]
    }

    set n 0
    set sigblank 0
    while {[set count [gets $fid line]] >= 0} {
        incr n
        if {$count == 0 && !$sigblank} continue
        # if only whitespace on line, also treat is as a blank
        if { [regexp {^[ 	]*$} $line] && !$sigblank } continue
 
        if [regexp {^[#]} $line] continue
        if $opt {
            regsub {^[	 ]+} $line {} line
            regsub {[ 	]+$} $line {} line
            regsub -all \[\ \	\]+$delim $line $delim line
            regsub -all $delim\[\ \	\]+ $line $delim line
            if {$opt == 1} {
                $vec add [split $line $delim]
            } else {
                $vec add "$n [split $line $delim]"
            }
        } else {
            $vec add $line
        }
    }
    close $fid
    return 1
}

proc dataFileWrite { file vec args } {

    if [catch {set fid [open $file w]} result] {
        return 0
    }

    set nocolon [expr [string first : $args]>=0]
    for {set i 0} {$i < [$vec size]} {incr i} {
        if [llength $args] {
            if $nocolon {
                set rawrec [$vec record $i]
                regsub -all : $rawrec . rec
            } else {
                set rec [$vec record $i]
            }
            puts $fid [eval format $args $rec]
        } else {
            puts $fid [$vec record $i]
        }
    }
    close $fid
    return 1
}

# This function is needed because 6.0.1 and earlier versions of hawk had a bug such
# that they wrote gwpasswd in wrong format (wrote out more fields than there were in gwpasswd.  
# Due to this bug, sgupdate breaks and we need to fix the format of gwpasswd.  

proc correctGwpasswd { sgdirectory } {

#Change format of gwpasswd
    if [file exists $sgdirectory] { 

        set historyfile [file join $sgdirectory pkhistory]
        set userfile [file join $sgdirectory gwpasswd]
        set usr(fieldnames)  {account cryptpass uid gid name lastchng minchng maxvalid warn expiry extra} 
        foreach x $usr(fieldnames)  {
            set usr($x) {}
        }

        vector users $usr(fieldnames)
        users key account uid
        if [dataFileRead $userfile users :] {

			dataUprev users [llength $usr(fieldnames)] {lastchng minchng maxvalid warn expiry extra} {{} {} {} {} {}} 1

            if ![dataFileWrite $userfile users %s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s] {
                #puts stdout "Error writing gwpasswd file
            }
                        
            set fileid [open $historyfile a]
			close $fileid
	    }
    }
}

proc checkProductLicense { path } {
	#Only need to worry about firewall.  Need not worry about RMC
	#The code guarantees that this function would not be called for the gui

    if [catch { set verstring [getVersionString $path] } result] {
        return 0
    }

    if { [string first "(invalid" $verstring] >=0 ||\
         [string first "(expired" $verstring] >=0 } {
        return 0
    } else {
        return 1
    }
}

proc getProductType { path prod } {

	if { [string compare $prod RMC] != 0 } {

		if [catch { set verstring [getVersionString $path] } result] {
			return "unknown"
		}

		if { [string first international $verstring] >= 0 } {
			return "international"
		} elseif { [string first french $verstring] >= 0 } {
			return "french"
		} else {
			return "domestic"
		}
	} else { 

	#should enter this piece only for RMC

	# This code needs to be added because with 6.0, version.exe was not installed
	# with RMC only installs.  This code can be removed later because after 6.0.1
	# version.exe would be installed with RMC only installs.  

	# In RMC only installs without version.exe, there is no direct way of finding 
	# the type of firewall.  It is not stored in any registry key.  The workaround is
	# that we check the product version of some known dll.  If it is 6.0 then it is domestic,
	# 6.0i is for international, else it is 6.0f
		
		set rmcPath [getInstalledProductPath RMC]
		regsub -all {\\} $rmcPath {\\\\} rmcPath
		set rmcPath $rmcPath\\\\mmcmixin.dll

		if [catch { set fileVersion [getFileVersion $rmcPath] } result] {		
			#puts stdout "Warning:  Unable to get file version info"
			return unknown
		}

		set stringLength [string length $fileVersion]
		set lastBit [string index $fileVersion [expr $stringLength-1]]
		if { [string compare $lastBit "i"] == 0 } {
			return international
		} elseif { [string compare $lastBit "f"] == 0 } {
			return french
		} else { 
			return domestic
		} 
	}
}


proc backoutPatch { restoredir } {

    set error 0
    disableInterrupts
    for { set i 0 } { $i < [restoreFiles size] } { incr i } {
        set src [file join [restoreFiles at $i srcfile]]
        set target [file join [restoreFiles at $i target]]

        if { [string compare $src "-"] == 0 } {
            file delete -force $target
        } else {
            if [catch { fileCopy $src $target} result] {
                incr error
                puts stdout "WARNING: unable to restore file [changeSlashes $target]..."
            }
        }
    }


    if !$error {
        file delete -force $restoredir
        puts stdout "Patch backed out... 
		puts stdout "If you have the firewall installed, then please reboot your machine now."
    } else {
        puts stdout "WARNING: Unable to back patch out completely... "
    }

    enableInterrupts
    exit 1
}

proc installPatch {} {
    
    clearScreen
    
    vector patches {product type low high new index}
    if ![dataFileRead Patch.ver patches] {
        puts stdout ""
        puts stdout "ERROR: Patch version information missing.  Make sure you are running"
        puts stdout "       the patch command from the top level of the unpacked patch tree."
        puts stdout ""
        exit 1
    }

  puts stdout "********************************************************"
  puts stdout "*                                                      *"
  puts stdout "* Please exit out of MMC before installing this patch. *"
  puts stdout "* To continue with patch installation, please enter y. *"
  puts stdout "* To abort, please enter a.                            *"
  puts stdout "*                                                      *"
  puts stdout "********************************************************"
  flush stdout;
  set response [gets stdin]
  set response [string tolower $response]
  
  if { [string compare $response "y"] != 0 } {

	puts stdout "You entered $response.  Exiting out of the patch process."
	puts stdout "Close MMC and then install the patch."
	exit 0
}

    puts stdout ""
    puts stdout "Checking for installed products..."
    puts stdout ""
    
    set prods [getInstalledProductList]
    
    set found 0
    set index -1
    set ctindex -1
    set newvers {}
    set oldvers {}
    set badprods {}
    
	#puts stdout "the length is [llength $prods]"
    for { set i 0 } { $i < [llength $prods] } { incr i } {
        set prod [lindex $prods $i]
		#puts stdout "the value of prod is $prod"
        if [catch { set path [getInstalledProductPath $prod] } result] {
			puts "cannot find installed product path for $prod"
            lappend badprods $i
            continue
        }
	#puts stdout "the path is $path"
        if [catch { set vers [getInstalledProductVersion $prod] } result] {
			puts stdout "cannot find product version for $prod"
            lappend badprods $i
            continue
        }
    
        if { [file exists $path] } {
            for { set j 0 } { $j < [patches size] } { incr j } {
                if { [string compare [patches at $j product] $prod] == 0 } {
					set v1 0; set v2 0; set v3 0
                    scan $vers "%d.%d.%d" v1 v2 v3
                    set  instver [expr 100 * $v1 + 10 * $v2 + $v3]
    
                    set v1 0; set v2 0; set v3 0
                    scan [patches at $j low] "%d.%d.%d" v1 v2 v3
                    set  low [expr 100 * $v1 + 10 * $v2 + $v3]
    
                    set v1 0; set v2 0; set v3 0
                    scan [patches at $j high] "%d.%d.%d" v1 v2 v3
                    set  high [expr 100 * $v1 + 10 * $v2 + $v3]
    
                    if { [string compare [patches at $j type] [getProductType $path $prod]] != 0 } {
                        # domestic/international mis-match, or unable to determine the right type
						#puts stdout "domestic international mismatch"
                        continue
                    }

                    if { $instver >= $low && $instver <= $high } {
                        incr found
                        set index $i
                        set ctindex $j
                        set newvers [patches at $j new]
                        set oldvers $vers
                    }
                }
            }
        } else {
		#puts stdout "$prod put in bad prod"
            lappend badprods $i
        }
    }

    set badprods [lsort -decreasing $badprods]
    
    foreach bad $badprods {
		#deletes the bad products from prods
        set prods [lreplace $prods $bad $bad]
    }
    
    if !$found {
        if [llength $prods] {
            puts stdout "Cannot find any products to patch..."
            puts stdout ""
            puts stdout "Patch found the following products installed: "
            foreach x $prods {
				set product [getProductType [getInstalledProductPath $x] $x]
                puts stdout "\t $product $x version [getInstalledProductVersion $x]"
            }
        } else {
            puts stdout "No installed Raptor products found! "
            puts stdout ""
            exit 1
        }
    
        puts stdout ""
        puts stdout "This patch applies to: "
        for { set j 0 } { $j < [patches size] } { incr j } {
            puts stdout "\t[patches at $j type] [patches at $j product] versions [patches at $j low] to [patches at $j high] (to version [patches at $j new])"
        }
        puts stdout ""
           
        exit 1
    }

    if { $found > 1 } {
        for { set i 0 } { $i < [llength $prods] } { incr i } {
			set prod [lindex $prods $i]
			installFiles $prod [patches at [patches find product $prod] index] $oldvers $newvers
		}
    } else {
		installFiles $prod [patches at [patches find product $prod] index] $oldvers $newvers
	}
    
	# While setting the product version, there are couple of restrictions placed
	# the way registrey keys are installed.  The restrictions are:
	# 1. If the firewall dlls are patched, then you need to update the patch version
	#	 of firewall and vpn registry keys.
	# 2. Where RMC key is installed depends on whether firewall is installed or not

	set firewall 0
	set RMC 0
	set RaptorRemote 0
	disableInterrupts
	for { set i 0 } {$i < [llength prods]} { incr i} {

		if { [string compare [lindex $prods $i] "firewall"] == 0} {
			set firewall 1
		} elseif { [string compare [lindex $prods $i] "RMC"] == 0} {
			set RMC 1
		} elseif { [string compare [lindex $prods $i] "RaptorRemote"] == 0} {
			set RaptorRemote 1
		}
	}

	puts stdout "Installing patch..."
    puts stdout ""

	if { [string compare $firewall 1] == 0} {
		# need to update the version of vpn and firewall
		setInstalledProductVersion firewall $newvers
		setInstalledProductVersion Vpn $newvers
	} else {
		# Note else state over here is important.  We need to update the RMC version
		# if it is RMC only install.  Also, RaptorRemote would not be there if firewall
		# is there. 

		if { [string compare $RMC 1] == 0} {
			#This is RMC only install.  Look for key at different place
			setInstalledProductVersion RMC $newvers
		}
		
		if { [string compare $RaptorRemote 1] == 0} {
		# need to update the version of vpn and RaptorRemote
			setInstalledProductVersion RaptorRemote $newvers
			setInstalledProductVersion Vpn $newvers
		}
		
	}

	enableInterrupts
}

proc installFiles { prod ctindex oldvers newvers } {

#  Need to change this to include multiple products
#   set prod [lindex $prods $index]
    puts stdout ""
    puts stdout "Ready to patch $prod from version $oldvers to version $newvers... "
    puts stdout ""


    set path [getInstalledProductPath $prod]
    if { [string compare $prod RMC] != 0 } {       
		#RMC does not care for the license key
        if ![checkProductLicense $path] { 
            puts stdout ""
            puts stdout "WARNING: The license key for this product is invalid or has expired."
            puts stdout "         The patch will still be applied, but the product will not function"
            puts stdout "         correctly until a valid license key is entered!"
            puts stdout ""
           
        }
    }

    set ctfile Contents.$ctindex

    puts stdout "Checking patch integrity for $prod ..."
    puts stdout ""
    
    vector patchFiles {srcfile mode uid gid target type fileversion checksum}
    if ![dataFileRead $ctfile patchFiles] {
        puts stdout ""
        puts stdout "ERROR: Patch contents file missing.  Make sure you are have "
        puts stdout "       downloaded and extracted the patch files correctly."
        puts stdout ""
        exit 1
    }

    set error 0
    for { set i 0 } { $i < [patchFiles size] } { incr i } {
        if { [string compare [file tail [patchFiles at $i srcfile]] "-"] != 0 } {
            if { [string compare [md5 [patchFiles at $i srcfile]] [patchFiles at $i checksum]] != 0 } {
                set error 1
            }
        }
    }
    
    if $error {
        puts stdout ""
        puts stdout "ERROR: Patch failed integrity check and thus will not be applied.  "
        puts stdout "       Please download a clean copy of the patch from the Raptor Systems web "
        puts stdout "       site or contact Raptor Systems Technical Support."
        exit 1
    }
	
	if { [string compare $prod "RMC"] != 0 } {

		puts stdout "Stopping $prod service...."
        puts stdout ""

        # shut down running services...
		if { [string compare $prod "firewall"] == 0 } {
			set shutdownCommand "Raptor Firewall"
		} elseif { [string compare $prod "RaptorRemote"] == 0 } {
			set shutdownCommand "RaptorRemote"
		}

        if [catch { shutdownService $shutdownCommand } result] {
            puts stdout "ERROR: Unable to shut down the $prod service.. Patch "
            puts stdout "       cannot be uninstalled!"
            puts stdout ""
            exit 1
		}
	}

	# check over here if the version given in the patchFiles vector is the same 
	# as the version on the machine.  Based on that delete the contents of the patchFiles

	for { set i 0 } { $i < [patchFiles size] } { incr i } {
		if { [string compare [patchFiles at $i fileversion] "-"] != 0 } {
			regsub -all {\\} $path {\\\\} srcfilepath
			set srcfile $srcfilepath\\\\[patchFiles at $i srcfile]
			set fileversion [getFileVersion $srcfile]

			if { $fileversion >= [patchFiles at $i fileversion] } {
				patchFiles delete $i				
			}
		}
	}


    # install files
	
    set instpath [getInstalledProductPath $prod]	
	if { [string compare $prod "RMC"] == 0 } {
		set instpath [file dirname $instpath]
	}

    set backupdir [file join $instpath patchsave]
    set restoredir [file join $instpath patchtmp]
    set backupcts [file join $backupdir Contents]
    set backupver [file join $backupdir version]
    set sgdirectory [file join $instpath sg]

    if ![file exists $backupdir] { file mkdir $backupdir }
    if ![file exists $restoredir] { file mkdir $restoredir }
    
    vector backupFiles {srcfile mode uid gid target type fileVersion checksum}
    dataFileRead $backupcts backupFiles
    vector restoreFiles {srcfile mode uid gid target type fileVersion checksum}

    # turn off control-c, other control signals
    disableInterrupts

	# We need to correct the gwpasswd format only for firewall or raptorremote.
	# We don't want to do it for RMC only install because there won't be any gwpasswd file.
	if { [string compare $prod "RMC"] != 0 } {
		correctGwpasswd $sgdirectory
    }

    for { set i 0 } { $i < [patchFiles size] } { incr i } {
        set newfile 0
        set usebackup 1
        set patchfile 0
        set src [file join [fileExpand [patchFiles at $i srcfile]]]
		if { [string compare $prod "RMC"] == 0 } {
			#deal patch uninstall files separately

			if { [string compare [patchFiles at $i type] "patch"] == 0} {
				set target [file join $instpath "patchsave" [patchFiles at $i target]]
			} else {
				set target [file join [getInstalledProductPath $prod] [patchFiles at $i target]]
			}

		} else {
			set target [file join [fileExpand [patchFiles at $i target]]]
		}

        set backup [file join $backupdir [file dirname $src] [file tail $target]]
        set relbackup [file join [file dirname $src] [file tail $target]]
        set restore [file join $restoredir [file tail $target]]

        # deleting a file
        if { [string compare [file tail $src] "-"] == 0 } {
            puts stdout "Removing file [changeSlashes $target]..."
            if ![file exists [file dirname $backup]] {
                 file mkdir [file dirname $backup]
            }

            if ![file exists $backup] {
                if [catch { fileCopy $target $backup } result] {
                    puts stdout ""
                    puts stdout "ERROR: Unable to backup file [changeSlashes $target]..."
                    puts stdout "       Backing out patch..."
                    puts stdout ""
                    backoutPatch $restoredir
                }

                set idx [backupFiles size]
                backupFiles add [patchFiles record $i]
                backupFiles at $idx srcfile $relbackup
                backupFiles at $idx target $target
            }

            if [catch { fileCopy $target $restore } result] {
                 puts stdout ""
                 puts stdout "ERROR: Unable to backup file [changeSlashes $target]..."
                 puts stdout "       Backing out patch..."
                 puts stdout ""
                 backoutPatch $restoredir
            }

            set idx [restoreFiles size]
            restoreFiles add [patchFiles record $i]
            restoreFiles at $idx srcfile $restore
            restoreFiles at $idx target $target

            file delete -force $target
            continue
        }

        if { [string compare [patchFiles at $i type] patch] == 0 } {
            set patchfile 1 
        } else {
            puts stdout "Patching file [changeSlashes $target]..."
        }

        if { !$patchfile && [file exists $target] } {
            if ![file exists $backup] {
                if ![file exists [file dirname $backup]] {
                     file mkdir [file dirname $backup]
                }
                if [catch { fileCopy $target $backup } result] {
                     puts stdout ""
                     puts stdout "ERROR: Unable to backup file [changeSlashes $target]..."
                     puts stdout "       Backing out patch..."
                     puts stdout ""
                     backoutPatch $restoredir
                }

                # only put expanded files into the backup contents
                set idx [backupFiles size]
                backupFiles add [patchFiles record $i]
                backupFiles at $idx srcfile $relbackup
                backupFiles at $idx target $target
            } else {
                set usebackup 0
                if [catch { fileCopy $target $restore } result] {
                     puts stdout ""
                     puts stdout "ERROR: Unable to backup file [changeSlashes $target]..."
                     puts stdout "       Backing out patch..."
                     puts stdout ""
                     backoutPatch $restoredir
                }
            }
        } else {
            set newfile 1

            # if this is a patch-associated file, don't add to files to remove
            if !$patchfile {
                set idx [backupFiles size]
                backupFiles add [patchFiles record $i]

                # add "delete this file" marker
                backupFiles at $idx srcfile "-"

                # only put expanded files into the backup contents
                backupFiles at $idx target $target
            }
        }
            
        if [catch { fileCopy $src $target } result] {
             puts stdout ""
             puts stdout "ERROR: Unable to copy file [changeSlashes $src] to [changeSlashes $target]..."
             puts stdout "       Backing out patch..."
             puts stdout ""
             backoutPatch $restoredir
        }

        set idx [restoreFiles size]
        if !$newfile {
            restoreFiles add [patchFiles record $i]
            restoreFiles at $idx srcfile $restore
            restoreFiles at $idx target $target

            if $usebackup {
                restoreFiles at $idx srcfile $backup
            }
        } else {
            restoreFiles add [patchFiles record $i]

            # add "delete this file" marker
            restoreFiles at $idx srcfile "-"
            restoreFiles at $idx target $target
        }
    }

    if ![dataFileWrite $backupcts backupFiles] {
        puts stdout ""
        puts stdout "ERROR: Unable to write uninstall data... "
        puts stdout "       Backing out patch..."
        puts stdout ""
        backoutPatch $restoredir
    }

    if ![file exists $backupver] {
        vector vvers { product version }
        vvers add [list $prod $oldvers]
        if { [string compare $prod RMC] != 0 } { 
            vvers add [list Vpn $oldvers]
        }

        if ![dataFileWrite $backupver vvers] {
            puts stdout ""
            puts stdout "ERROR: Unable to write uninstall data... "
            puts stdout "       Backing out patch..."
            puts stdout ""
            backoutPatch $restoredir
        }

        vvers destroy
    }

    # done with install... Clean up restore directory
    file delete -force $restoredir

	
    enableInterrupts

    restoreFiles destroy
    backupFiles destroy
    patchFiles destroy	
}

proc uninstallPatch {} {

  puts stdout "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
  puts stdout "!                                                       !"
  puts stdout "! Please exit out of MMC before uninstalling this patch.!"
  puts stdout "! To continue with patch uninstall, please enter y.     !"
  puts stdout "! To abort patch uninstall, please enter a.             !"
  puts stdout "!                                                       !"
  puts stdout "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
  flush stdout;
  set response [gets stdin]
  set response [string tolower $response]
  
  if { [string compare $response "y"] != 0 } {

	puts stdout "You entered $response.  Exiting out of the patch uninstall process."
	puts stdout "Close MMC and then uninstall the patch."
	exit 0
}

    catch { vvers destroy }
    vector vvers { product version }
    if { ![dataFileRead version vvers] || [vvers size] == 0 } {
        puts stdout ""
        puts stdout "ERROR: Unable to find uninstall data.. Please run the"
        puts stdout "       patch uninstall from the patch save directory!"
        puts stdout ""
        exit 1
    }

    catch { backupFiles destroy }
    vector backupFiles {srcfile mode uid gid target type fileVersion checksum}

    if { ![dataFileRead Contents backupFiles] || [backupFiles size] == 0 } {
        puts stdout ""
        puts stdout "ERROR: Unable to find list of files to uninstall.  Please"
        puts stdout "       run the patch uninstall from the patch save directory!"
        puts stdout ""
        exit 1
    }

    set prod [vvers at 0 product]
    if { [string compare $prod RMC] != 0 } {
        puts stdout "Stopping $prod service...."
        puts stdout ""

        # shut down running services...
		if { [string compare $prod "firewall"] == 0 } {
			set shutdownCommand "Raptor Firewall"
		} elseif { [string compare $prod "RaptorRemote"] == 0 } {
			set shutdownCommand "RaptorRemote"
		}

        if [catch { shutdownService $shutdownCommand } result] {
            puts stdout "ERROR: Unable to shut down the $prod service.. Patch "
            puts stdout "       cannot be uninstalled!"
            puts stdout ""
            exit 1
        }
    }

    disableInterrupts

    set error 0
    for { set i 0 } { $i < [backupFiles size] } { incr i } {
        set src [file join [backupFiles at $i srcfile]]
        set target [file join [backupFiles at $i target]]

        # if delete marker, delete instead of 
        if { [string compare $src "-"] == 0 } {
            puts stdout "Removing file [changeSlashes $target]..."
            file delete -force $target
        } else {
            puts stdout "Restoring file [changeSlashes $target]..."
            if [catch { fileCopy $src $target } result] {
                incr error
                puts stdout ""
                puts stdout "WARNING: unable to restore file [changeSlashes $target]: $result"
            } else {
                file delete -force $src
            }
        }
    }

    if !$error {
        for { set i 0 } { $i < [vvers size] } { incr i } {
            setInstalledProductVersion [vvers at $i product] [vvers at $i version]
        }
        file delete -force version
        file delete -force Contents
    } else {
        puts stdout ""
        puts stdout "WARNING: Patch uninstall did not complete succesfully."
        puts stdout ""
        exit 1
    }

    enableInterrupts
}

### code starts here ####
global uninstall

if !$uninstall {
    installPatch
    puts stdout ""
    puts stdout "Patch install completed succesfully!"
    puts stdout "Please reboot your machine if you have the firewall installed."
    puts stdout ""
} else {
    uninstallPatch
    puts stdout ""
    puts stdout "Patch uninstall completed succesfully!  Please reboot your machine."
    puts stdout ""
}

